@@ -10,39 +10,21 @@ PyInt_GetMax(void)
 	return LONG_MAX;	/* To initialize sys.maxint */
 }
 
-/* Standard Booleans */
-
-PyIntObject _Py_ZeroStruct = {
-	PyObject_HEAD_INIT(&PyInt_Type)
-	0
-};
-
-PyIntObject _Py_TrueStruct = {
-	PyObject_HEAD_INIT(&PyInt_Type)
-	1
-};
-
-/* Return 1 if exception raised, 0 if caller should retry using longs */
-static int
-err_ovf(char *msg)
-{
-	if (PyErr_Warn(PyExc_OverflowWarning, msg) < 0) {
-		if (PyErr_ExceptionMatches(PyExc_OverflowWarning))
-			PyErr_SetString(PyExc_OverflowError, msg);
-		return 1;
-	}
-	else
-		return 0;
-}
-
 /* Integers are quite normal objects, to make object handling uniform.
    (Using odd pointers to represent integers would save much space
    but require extra checks for this special case throughout the code.)
-   Since, a typical Python program spends much of its time allocating
+   Since a typical Python program spends much of its time allocating
    and deallocating integers, these operations should be very fast.
    Therefore we use a dedicated allocation scheme with a much lower
    overhead (in space and time) than straight malloc(): a simple
    dedicated free list, filled when necessary with memory from malloc().
+
+   block_list is a singly-linked list of all PyIntBlocks ever allocated,
+   linked via their next members.  PyIntBlocks are never returned to the
+   system before shutdown (PyInt_Fini).
+
+   free_list is a singly-linked list of available PyIntObjects, linked
+   via abuse of their ob_type members.
 */
 
 #define BLOCK_SIZE	1000	/* 1K less typical malloc overhead */
@@ -63,12 +45,14 @@ static PyIntObject *
 fill_free_list(void)
 {
 	PyIntObject *p, *q;
-	/* XXX Int blocks escape the object heap. Use PyObject_MALLOC ??? */
+	/* Python's object allocator isn't appropriate for large blocks. */
 	p = (PyIntObject *) PyMem_MALLOC(sizeof(PyIntBlock));
 	if (p == NULL)
 		return (PyIntObject *) PyErr_NoMemory();
 	((PyIntBlock *)p)->next = block_list;
 	block_list = (PyIntBlock *)p;
+	/* Link the int objects together, from rear to front, then return
+	   the address of the last int object in the block. */
 	p = &((PyIntBlock *)p)->objects[0];
 	q = p + N_INTOBJECTS;
 	while (--q > p)
@@ -81,7 +65,7 @@ fill_free_list(void)
 #define NSMALLPOSINTS		100
 #endif
 #ifndef NSMALLNEGINTS
-#define NSMALLNEGINTS		1
+#define NSMALLNEGINTS		5
 #endif
 #if NSMALLNEGINTS + NSMALLPOSINTS > 0
 /* References to small integers are saved in this array so that they
@@ -100,8 +84,8 @@ PyInt_FromLong(long ival)
 {
 	register PyIntObject *v;
 #if NSMALLNEGINTS + NSMALLPOSINTS > 0
-	if (-NSMALLNEGINTS <= ival && ival < NSMALLPOSINTS &&
-	    (v = small_ints[ival + NSMALLNEGINTS]) != NULL) {
+	if (-NSMALLNEGINTS <= ival && ival < NSMALLPOSINTS) {
+		v = small_ints[ival + NSMALLNEGINTS];
 		Py_INCREF(v);
 #ifdef COUNT_ALLOCS
 		if (ival >= 0)
@@ -116,18 +100,11 @@ PyInt_FromLong(long ival)
 		if ((free_list = fill_free_list()) == NULL)
 			return NULL;
 	}
-	/* PyObject_New is inlined */
+	/* Inline PyObject_New */
 	v = free_list;
 	free_list = (PyIntObject *)v->ob_type;
 	PyObject_INIT(v, &PyInt_Type);
 	v->ob_ival = ival;
-#if NSMALLNEGINTS + NSMALLPOSINTS > 0
-	if (-NSMALLNEGINTS <= ival && ival < NSMALLPOSINTS) {
-		/* save this one for a following allocation */
-		Py_INCREF(v);
-		small_ints[ival + NSMALLNEGINTS] = v;
-	}
-#endif
 	return (PyObject *) v;
 }
 
@@ -169,9 +146,21 @@ PyInt_AsLong(register PyObject *op)
 	if (io == NULL)
 		return -1;
 	if (!PyInt_Check(io)) {
-		PyErr_SetString(PyExc_TypeError,
-				"nb_int should return int object");
-		return -1;
+		if (PyLong_Check(io)) {
+			/* got a long? => retry int conversion */
+			val = PyLong_AsLong((PyObject *)io);
+			Py_DECREF(io);
+			if ((val == -1) && PyErr_Occurred())
+				return -1;
+			return val;
+		}
+		else
+		{
+			Py_DECREF(io);
+			PyErr_SetString(PyExc_TypeError,
+					"nb_int should return int object");
+			return -1;
+		}
 	}
 
 	val = PyInt_AS_LONG(io);
@@ -180,6 +169,96 @@ PyInt_AsLong(register PyObject *op)
 	return val;
 }
 
+unsigned long
+PyInt_AsUnsignedLongMask(register PyObject *op)
+{
+	PyNumberMethods *nb;
+	PyIntObject *io;
+	unsigned long val;
+
+	if (op && PyInt_Check(op))
+		return PyInt_AS_LONG((PyIntObject*) op);
+	if (op && PyLong_Check(op))
+		return PyLong_AsUnsignedLongMask(op);
+
+	if (op == NULL || (nb = op->ob_type->tp_as_number) == NULL ||
+	    nb->nb_int == NULL) {
+		PyErr_SetString(PyExc_TypeError, "an integer is required");
+		return -1;
+	}
+
+	io = (PyIntObject*) (*nb->nb_int) (op);
+	if (io == NULL)
+		return -1;
+	if (!PyInt_Check(io)) {
+		if (PyLong_Check(io)) {
+			val = PyLong_AsUnsignedLongMask((PyObject *)io);
+			Py_DECREF(io);
+			if (PyErr_Occurred())
+				return -1;
+			return val;
+		}
+		else
+		{
+			Py_DECREF(io);
+			PyErr_SetString(PyExc_TypeError,
+					"nb_int should return int object");
+			return -1;
+		}
+	}
+
+	val = PyInt_AS_LONG(io);
+	Py_DECREF(io);
+
+	return val;
+}
+
+#ifdef HAVE_LONG_LONG
+unsigned PY_LONG_LONG
+PyInt_AsUnsignedLongLongMask(register PyObject *op)
+{
+	PyNumberMethods *nb;
+	PyIntObject *io;
+	unsigned PY_LONG_LONG val;
+
+	if (op && PyInt_Check(op))
+		return PyInt_AS_LONG((PyIntObject*) op);
+	if (op && PyLong_Check(op))
+		return PyLong_AsUnsignedLongLongMask(op);
+
+	if (op == NULL || (nb = op->ob_type->tp_as_number) == NULL ||
+	    nb->nb_int == NULL) {
+		PyErr_SetString(PyExc_TypeError, "an integer is required");
+		return -1;
+	}
+
+	io = (PyIntObject*) (*nb->nb_int) (op);
+	if (io == NULL)
+		return -1;
+	if (!PyInt_Check(io)) {
+		if (PyLong_Check(io)) {
+			val = PyLong_AsUnsignedLongLongMask((PyObject *)io);
+			Py_DECREF(io);
+			if (PyErr_Occurred())
+				return -1;
+			return val;
+		}
+		else
+		{
+			Py_DECREF(io);
+			PyErr_SetString(PyExc_TypeError,
+					"nb_int should return int object");
+			return -1;
+		}
+	}
+
+	val = PyInt_AS_LONG(io);
+	Py_DECREF(io);
+
+	return val;
+}
+#endif
+
 PyObject *
 PyInt_FromString(char *s, char **pend, int base)
 {
@@ -188,15 +267,19 @@ PyInt_FromString(char *s, char **pend, int base)
 	char buffer[256]; /* For errors */
 
 	if ((base != 0 && base < 2) || base > 36) {
-		PyErr_SetString(PyExc_ValueError, "int() base must be >= 2 and <= 36");
+		PyErr_SetString(PyExc_ValueError,
+				"int() base must be >= 2 and <= 36");
 		return NULL;
 	}
 
 	while (*s && isspace(Py_CHARMASK(*s)))
 		s++;
 	errno = 0;
-	if (base == 0 && s[0] == '0')
+	if (base == 0 && s[0] == '0') {
 		x = (long) PyOS_strtoul(s, &end, base);
+		if (x < 0)
+			return PyLong_FromString(s, pend, base);
+	}
 	else
 		x = PyOS_strtol(s, &end, base);
 	if (end == s || !isalnum(Py_CHARMASK(end[-1])))
@@ -210,12 +293,8 @@ PyInt_FromString(char *s, char **pend, int base)
 		PyErr_SetString(PyExc_ValueError, buffer);
 		return NULL;
 	}
-	else if (errno != 0) {
-		PyOS_snprintf(buffer, sizeof(buffer),
-			      "int() literal too large: %.200s", s);
-		PyErr_SetString(PyExc_ValueError, buffer);
-		return NULL;
-	}
+	else if (errno != 0)
+		return PyLong_FromString(s, pend, base);
 	if (pend)
 		*pend = end;
 	return PyInt_FromLong(x);
@@ -225,16 +304,19 @@ PyInt_FromString(char *s, char **pend, int base)
 PyObject *
 PyInt_FromUnicode(Py_UNICODE *s, int length, int base)
 {
-	char buffer[256];
+	PyObject *result;
+	char *buffer = PyMem_MALLOC(length+1);
 
-	if (length >= sizeof(buffer)) {
-		PyErr_SetString(PyExc_ValueError,
-				"int() literal too large to convert");
+	if (buffer == NULL)
+		return NULL;
+
+	if (PyUnicode_EncodeDecimal(s, length, buffer, NULL)) {
+		PyMem_FREE(buffer);
 		return NULL;
 	}
-	if (PyUnicode_EncodeDecimal(s, length, buffer, NULL))
-		return NULL;
-	return PyInt_FromString(buffer, NULL, base);
+	result = PyInt_FromString(buffer, NULL, base);
+	PyMem_FREE(buffer);
+	return result;
 }
 #endif
 
@@ -298,8 +380,6 @@ int_add(PyIntObject *v, PyIntObject *w)
 	x = a + b;
 	if ((x^a) >= 0 || (x^b) >= 0)
 		return PyInt_FromLong(x);
-	if (err_ovf("integer addition"))
-		return NULL;
 	return PyLong_Type.tp_as_number->nb_add((PyObject *)v, (PyObject *)w);
 }
 
@@ -312,8 +392,6 @@ int_sub(PyIntObject *v, PyIntObject *w)
 	x = a - b;
 	if ((x^a) >= 0 || (x^~b) >= 0)
 		return PyInt_FromLong(x);
-	if (err_ovf("integer subtraction"))
-		return NULL;
 	return PyLong_Type.tp_as_number->nb_subtract((PyObject *)v,
 						     (PyObject *)w);
 }
@@ -344,14 +422,6 @@ one that can lose catastrophic amounts of information, it's the native long
 product that must have overflowed.
 */
 
-/* Return true if the sq_repeat method should be used */
-#define USE_SQ_REPEAT(o) (!PyInt_Check(o) && \
-			  o->ob_type->tp_as_sequence && \
-			  o->ob_type->tp_as_sequence->sq_repeat && \
-			  !(o->ob_type->tp_as_number && \
-                            o->ob_type->tp_flags & Py_TPFLAGS_CHECKTYPES && \
-			    o->ob_type->tp_as_number->nb_multiply))
-
 static PyObject *
 int_mul(PyObject *v, PyObject *w)
 {
@@ -360,44 +430,6 @@ int_mul(PyObject *v, PyObject *w)
 	double doubled_longprod;	/* (double)longprod */
 	double doubleprod;		/* (double)a * (double)b */
 
-	if (USE_SQ_REPEAT(v)) {
-	  repeat:
-		/* sequence * int */
-		a = PyInt_AsLong(w);
-#if LONG_MAX != INT_MAX
-		if (a > INT_MAX) {
-			PyErr_SetString(PyExc_ValueError,
-					"sequence repeat count too large");
-			return NULL;
-		}
-		else if (a < INT_MIN)
-			a = INT_MIN;
-		/* XXX Why don't I either
-
-		   - set a to -1 whenever it's negative (after all,
-		     sequence repeat usually treats negative numbers
-		     as zero(); or
-
-		   - raise an exception when it's less than INT_MIN?
-
-		   I'm thinking about a hypothetical use case where some
-		   sequence type might use a negative value as a flag of
-		   some kind.  In those cases I don't want to break the
-		   code by mapping all negative values to -1.  But I also
-		   don't want to break e.g. []*(-sys.maxint), which is
-		   perfectly safe, returning [].  As a compromise, I do
-		   map out-of-range negative values.
-		*/
-#endif
-		return (*v->ob_type->tp_as_sequence->sq_repeat)(v, a);
-	}
-	if (USE_SQ_REPEAT(w)) {
-		PyObject *tmp = v;
-		v = w;
-		w = tmp;
-		goto repeat;
-	}
-
 	CONVERT_TO_LONG(v, a);
 	CONVERT_TO_LONG(w, b);
 	longprod = a * b;
@@ -423,13 +455,22 @@ int_mul(PyObject *v, PyObject *w)
 		   32 * absdiff <= absprod -- 5 good bits is "close enough" */
 		if (32.0 * absdiff <= absprod)
 			return PyInt_FromLong(longprod);
-		else if (err_ovf("integer multiplication"))
-			return NULL;
 		else
 			return PyLong_Type.tp_as_number->nb_multiply(v, w);
 	}
 }
 
+/* Integer overflow checking for unary negation: on a 2's-complement
+ * box, -x overflows iff x is the most negative long.  In this case we
+ * get -x == x.  However, -x is undefined (by C) if x /is/ the most
+ * negative long (it's a signed overflow case), and some compilers care.
+ * So we cast x to unsigned long first.  However, then other compilers
+ * warn about applying unary minus to an unsigned operand.  Hence the
+ * weird "0-".
+ */
+#define UNARY_NEG_WOULD_OVERFLOW(x)	\
+	((x) < 0 && (unsigned long)(x) == 0-(unsigned long)(x))
+
 /* Return type of i_divmod */
 enum divmod_result {
 	DIVMOD_OK,		/* Correct result */
@@ -449,11 +490,8 @@ i_divmod(register long x, register long y,
 		return DIVMOD_ERROR;
 	}
 	/* (-sys.maxint-1)/-1 is the only overflow case. */
-	if (y == -1 && x < 0 && x == -x) {
-		if (err_ovf("integer division"))
-			return DIVMOD_ERROR;
+	if (y == -1 && UNARY_NEG_WOULD_OVERFLOW(x))
 		return DIVMOD_OVERFLOW;
-	}
 	xdivy = x / y;
 	xmody = x - xdivy * y;
 	/* If the signs of x and y differ, and the remainder is non-0,
@@ -602,8 +640,6 @@ int_pow(PyIntObject *v, PyIntObject *w, PyIntObject *z)
 			if (temp == 0)
 				break; /* Avoid ix / 0 */
 			if (ix / temp != prev) {
-				if (err_ovf("integer exponentiation"))
-					return NULL;
 				return PyLong_Type.tp_as_number->nb_power(
 					(PyObject *)v,
 					(PyObject *)w,
@@ -614,9 +650,7 @@ int_pow(PyIntObject *v, PyIntObject *w, PyIntObject *z)
 	        if (iw==0) break;
 	 	prev = temp;
 	 	temp *= temp;	/* Square the value of temp */
-	 	if (prev!=0 && temp/prev!=prev) {
-			if (err_ovf("integer exponentiation"))
-				return NULL;
+	 	if (prev != 0 && temp / prev != prev) {
 			return PyLong_Type.tp_as_number->nb_power(
 				(PyObject *)v, (PyObject *)w, (PyObject *)z);
 		}
@@ -645,14 +679,11 @@ int_pow(PyIntObject *v, PyIntObject *w, PyIntObject *z)
 static PyObject *
 int_neg(PyIntObject *v)
 {
-	register long a, x;
+	register long a;
 	a = v->ob_ival;
-	x = -a;
-	if (a < 0 && x < 0) {
-		PyObject *o;
-		if (err_ovf("integer negation"))
-			return NULL;
-		o = PyLong_FromLong(a);
+        /* check for overflow */
+	if (UNARY_NEG_WOULD_OVERFLOW(a)) {
+		PyObject *o = PyLong_FromLong(a);
 		if (o != NULL) {
 			PyObject *result = PyNumber_Negative(o);
 			Py_DECREF(o);
@@ -660,7 +691,7 @@ int_neg(PyIntObject *v)
 		}
 		return NULL;
 	}
-	return PyInt_FromLong(x);
+	return PyInt_FromLong(-a);
 }
 
 static PyObject *
@@ -698,7 +729,9 @@ int_invert(PyIntObject *v)
 static PyObject *
 int_lshift(PyIntObject *v, PyIntObject *w)
 {
-	register long a, b;
+	long a, b, c;
+	PyObject *vv, *ww, *result;
+
 	CONVERT_TO_LONG(v, a);
 	CONVERT_TO_LONG(w, b);
 	if (b < 0) {
@@ -708,10 +741,35 @@ int_lshift(PyIntObject *v, PyIntObject *w)
 	if (a == 0 || b == 0)
 		return int_pos(v);
 	if (b >= LONG_BIT) {
-		return PyInt_FromLong(0L);
+		vv = PyLong_FromLong(PyInt_AS_LONG(v));
+		if (vv == NULL)
+			return NULL;
+		ww = PyLong_FromLong(PyInt_AS_LONG(w));
+		if (ww == NULL) {
+			Py_DECREF(vv);
+			return NULL;
+		}
+		result = PyNumber_Lshift(vv, ww);
+		Py_DECREF(vv);
+		Py_DECREF(ww);
+		return result;
 	}
-	a = (long)((unsigned long)a << b);
-	return PyInt_FromLong(a);
+	c = a << b;
+	if (a != Py_ARITHMETIC_RIGHT_SHIFT(long, c, b)) {
+		vv = PyLong_FromLong(PyInt_AS_LONG(v));
+		if (vv == NULL)
+			return NULL;
+		ww = PyLong_FromLong(PyInt_AS_LONG(w));
+		if (ww == NULL) {
+			Py_DECREF(vv);
+			return NULL;
+		}
+		result = PyNumber_Lshift(vv, ww);
+		Py_DECREF(vv);
+		Py_DECREF(ww);
+		return result;
+	}
+	return PyInt_FromLong(c);
 }
 
 static PyObject *
@@ -800,7 +858,9 @@ int_oct(PyIntObject *v)
 {
 	char buf[100];
 	long x = v -> ob_ival;
-	if (x == 0)
+	if (x < 0)
+		PyOS_snprintf(buf, sizeof(buf), "-0%lo", -x);
+	else if (x == 0)
 		strcpy(buf, "0");
 	else
 		PyOS_snprintf(buf, sizeof(buf), "0%lo", x);
@@ -812,11 +872,14 @@ int_hex(PyIntObject *v)
 {
 	char buf[100];
 	long x = v -> ob_ival;
-	PyOS_snprintf(buf, sizeof(buf), "0x%lx", x);
+	if (x < 0)
+		PyOS_snprintf(buf, sizeof(buf), "-0x%lx", -x);
+	else
+		PyOS_snprintf(buf, sizeof(buf), "0x%lx", x);
 	return PyString_FromString(buf);
 }
 
-staticforward PyObject *
+static PyObject *
 int_subtype_new(PyTypeObject *type, PyObject *args, PyObject *kwds);
 
 static PyObject *
@@ -857,28 +920,52 @@ static PyObject *
 int_subtype_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
 	PyObject *tmp, *new;
+	long ival;
 
 	assert(PyType_IsSubtype(type, &PyInt_Type));
 	tmp = int_new(&PyInt_Type, args, kwds);
 	if (tmp == NULL)
 		return NULL;
-	assert(PyInt_Check(tmp));
+	if (!PyInt_Check(tmp)) {
+		ival = PyLong_AsLong(tmp);
+		if (ival == -1 && PyErr_Occurred()) {
+			Py_DECREF(tmp);
+			return NULL;
+		}
+	} else {
+		ival = ((PyIntObject *)tmp)->ob_ival;
+	}
+
 	new = type->tp_alloc(type, 0);
-	if (new == NULL)
+	if (new == NULL) {
+		Py_DECREF(tmp);
 		return NULL;
-	((PyIntObject *)new)->ob_ival = ((PyIntObject *)tmp)->ob_ival;
+	}
+	((PyIntObject *)new)->ob_ival = ival;
 	Py_DECREF(tmp);
 	return new;
 }
 
-static char int_doc[] =
+static PyObject *
+int_getnewargs(PyIntObject *v)
+{
+	return Py_BuildValue("(l)", v->ob_ival);
+}
+
+static PyMethodDef int_methods[] = {
+	{"__getnewargs__",	(PyCFunction)int_getnewargs,	METH_NOARGS},
+	{NULL,		NULL}		/* sentinel */
+};
+
+PyDoc_STRVAR(int_doc,
 "int(x[, base]) -> integer\n\
 \n\
 Convert a string or number to an integer, if possible.  A floating point\n\
 argument will be truncated towards zero (this does not include a string\n\
 representation of a floating point number!)  When converting a string, use\n\
 the optional base.  It is an error to supply a base when converting a\n\
-non-string.";
+non-string. If the argument is outside the integer range a long object\n\
+will be returned instead.");
 
 static PyNumberMethods int_as_number = {
 	(binaryfunc)int_add,	/*nb_add*/
@@ -951,7 +1038,7 @@ PyTypeObject PyInt_Type = {
 	0,					/* tp_weaklistoffset */
 	0,					/* tp_iter */
 	0,					/* tp_iternext */
-	0,					/* tp_methods */
+	int_methods,				/* tp_methods */
 	0,					/* tp_members */
 	0,					/* tp_getset */
 	0,					/* tp_base */
@@ -962,9 +1049,29 @@ PyTypeObject PyInt_Type = {
 	0,					/* tp_init */
 	0,					/* tp_alloc */
 	int_new,				/* tp_new */
-	(destructor)int_free,         		/* tp_free */
+	(freefunc)int_free,           		/* tp_free */
 };
 
+int
+_PyInt_Init(void)
+{
+	PyIntObject *v;
+	int ival;
+#if NSMALLNEGINTS + NSMALLPOSINTS > 0
+	for (ival = -NSMALLNEGINTS; ival < NSMALLPOSINTS; ival++) {
+              if (!free_list && (free_list = fill_free_list()) == NULL)
+			return 0;
+		/* PyObject_New is inlined */
+		v = free_list;
+		free_list = (PyIntObject *)v->ob_type;
+		PyObject_INIT(v, &PyInt_Type);
+		v->ob_ival = ival;
+		small_ints[ival + NSMALLNEGINTS] = v;
+	}
+#endif
+	return 1;
+}
+
 void
 PyInt_Fini(void)
 {
@@ -1025,7 +1132,7 @@ PyInt_Fini(void)
 			}
 		}
 		else {
-			PyMem_FREE(list); /* XXX PyObject_FREE ??? */
+			PyMem_FREE(list);
 			bf++;
 		}
 		isum += irem;
